/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gcc-compat.h"
#include "para.h"
#include <ortp/ortp.h>

#include "ortp.h"
#include "recv.h"
#include "ortp_recv.cmdline.h"


#include "error.h"
#include "audiod.h"
#define CHUNK_SIZE 128 * 1024

extern gint msg_to_buf(mblk_t *, char *, gint);



struct private_ortp_recv_data {
	RtpSession *session;
	struct timeval start, chunk_tv, timeout, next_chunk;
	int c_bad; /* number consecutive bad chunks */
	guint32 timestamp, chunk_ts;
	struct gengetopt_args_info *conf;
	int need_header;
};

#if 1
static void recalc_timestamp(struct receiver_node *rn)
{
	struct timeval now, tmp, min_delay = {0, 5000};
	struct private_ortp_recv_data *pord = rn->private_data;

	gettimeofday(&now, NULL);
	tv_add(&now, &min_delay, &pord->next_chunk);
	if (pord->start.tv_sec) {
		int new_ts;

		tv_diff(&now, &pord->start, &tmp);
		new_ts = rtp_session_time_to_ts(pord->session, tv2ms(&tmp));
//		if (pord->c_bad > 50)
			pord->timestamp = new_ts;
	} else
		pord->timestamp = 0;
	PARA_DEBUG_LOG("ts=%u, c_bad = %d\n", pord->timestamp, pord->c_bad);

}
#endif

static int ortp_recv_pre_select(struct receiver_node *rn,
		__unused fd_set *rfds, __unused fd_set *wfds,
		struct timeval *timeout)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	struct timeval now, tmp;

	gettimeofday(&now, NULL);
	if (tv_diff(&now, &pord->next_chunk, &tmp) < 0)
		pord->timeout = tmp;
	else {
		pord->timeout.tv_sec = 0;
		pord->timeout.tv_usec = 1000;
	}
	if (tv_diff(timeout, &pord->timeout, NULL) > 0)
		*timeout = pord->timeout;
	return -1; /* we did not modify the fd sets */
}

static void compute_next_chunk(struct timeval *now,
		struct private_ortp_recv_data *pord)
{
	if (!pord->start.tv_sec) {
		pord->start = *now;
		tv_add(&pord->chunk_tv, now, &pord->next_chunk);
	} else {
		struct timeval tmp;
		tv_add(&pord->chunk_tv, &pord->next_chunk, &tmp);
		pord->next_chunk = tmp;
	}
	pord->timestamp += pord->chunk_ts;
	PARA_DEBUG_LOG("next chunk (ts = %d) due at %lu:%lu\n",
		pord->timestamp, pord->next_chunk.tv_sec,
		pord->next_chunk.tv_usec);
}

int ortp_recv_post_select(struct receiver_node *rn, int select_ret,
	__unused fd_set *rfds, __unused fd_set *wfds)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	mblk_t *mp;
	int ret;
	char tmpbuf[CHUNK_SIZE + 3];
	struct timeval now;

	gettimeofday(&now, NULL);
	PARA_DEBUG_LOG("rn: %p, pord: %p, session: %p\n", rn, pord, pord->session);
	if (pord->start.tv_sec) {
		struct timeval diff;
		if (tv_diff(&now, &pord->next_chunk, &diff) < 0) {
			long unsigned diff_ms = tv2ms(&diff);
			if (diff_ms > 5) {
				PARA_DEBUG_LOG("too early: %lu ms left\n",
					diff_ms);
				return 1;
			}
		}
	}
	mp = rtp_session_recvm_with_ts(pord->session, pord->timestamp);
	if (!mp) {
		pord->c_bad++;
		if ((pord->c_bad > 500 && pord->start.tv_sec) || pord->c_bad > 5000)
			return -E_TOO_MANY_BAD_CHUNKS;
//		PARA_NOTICE_LOG("did not receive buffer, ts=%u, c_bad = %d, to = %lu\n",
//			pord->timestamp, pord->c_bad, tv2ms(&tmp));
		if (pord->c_bad > 100)
			recalc_timestamp(rn);
		return 1;
	}
	/* okay, we have a chunk of data */
	PARA_DEBUG_LOG("have it ts = %d, chunk_ts = %d, loaded: %d, bad: %d\n",
		pord->timestamp, pord->chunk_ts, rn->loaded, pord->c_bad);
	if (!pord->start.tv_sec)
		pord->start = now;
	ret = msg_to_buf(mp, tmpbuf, CHUNK_SIZE);
	if (ret <= 0) {
		ret = -E_MSG_TO_BUF;
		goto err_out;
	}
	switch (tmpbuf[0]) {
	unsigned header_len, payload_len;
	case ORTP_EOF:
		ret = 0;
		goto err_out;
	case ORTP_BOF:
		PARA_INFO_LOG("bof (%d)\n", ret);
		pord->need_header = 0;
		/* fall through */
	case ORTP_DATA:
//		PARA_INFO_LOG("data (%d)\n", ret);
		if (pord->need_header) /* can't use the data, wait for header */
			goto success;
		if (ret + rn->loaded >= CHUNK_SIZE + 3) {
			ret = -E_OVERRUN;
			goto err_out;
		}
		if (ret > 4) {
			memcpy(rn->buf + rn->loaded, tmpbuf + 3, ret - 3);
			rn->loaded += ret - 3;
		}
		goto success;
	case ORTP_HEADER:
		PARA_DEBUG_LOG("header (%d)\n", ret);
		if (ret < 4) {
			ret = -E_SHORT_HEADER;
			goto err_out;
		}
		if (pord->need_header) {
			pord->need_header = 0;
			memcpy(rn->buf, tmpbuf + 3, ret - 3);
			rn->loaded += ret - 3;
			goto success;
		}
		header_len =  ((unsigned)tmpbuf[2] & 0xff) +  256 * ((unsigned)tmpbuf[1] & 0xff);
		if (header_len  + 3 > ret) {
			ret = -E_INVALID_HEADER;
			goto err_out;
		}
		payload_len = ret - 3 - header_len;
//		PARA_INFO_LOG("len: %d header_len: %d, payload_len: %d, loaded: %d\n", ret,
//			header_len, payload_len, rn->loaded);
		if (rn->loaded + payload_len > CHUNK_SIZE) {
			ret = -E_OVERRUN;
			goto err_out;
		}
		if (payload_len)
			memcpy(rn->buf + rn->loaded, tmpbuf
				 + (ret - payload_len), payload_len);
		rn->loaded += payload_len;
		goto success;
	}
success:
	freemsg(mp);
	pord->c_bad = 0;
	compute_next_chunk(&now, pord);
	return 1;
err_out:
	freemsg(mp);
	return ret;
}

static void ortp_shutdown(void)
{
//	ortp_global_stats_display();
	ortp_exit();
}

static void ortp_recv_close(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord = rn->private_data;

	rtp_session_destroy(pord->session);
	free(rn->private_data);
	free(rn->buf);
}

static void *ortp_recv_parse_config(int argc, char **argv)
{
	int ret;

	struct gengetopt_args_info *tmp = para_calloc(sizeof(struct gengetopt_args_info));

	ret = ortp_recv_cmdline_parser(argc, argv, tmp)? -E_ORTP_SYNTAX : 1;
	if (ret > 0)
		return tmp;
	free(tmp);
	return NULL;
}

static int ortp_recv_open(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord;

	rn->buf = para_calloc(CHUNK_SIZE);

	rn->private_data = para_calloc(sizeof(struct private_ortp_recv_data));
	pord = rn->private_data;
	pord->session = rtp_session_new(RTP_SESSION_RECVONLY);
	PARA_DEBUG_LOG("receiver node: %p, pord: %p\n", rn, pord);
	pord->conf = rn->conf;
//	rtp_session_set_jitter_compensation(pord->session, 900);
	pord->need_header = pord->conf->header_given;
	pord->chunk_tv.tv_sec = 0;
	if (!rn->chunk_time || pord->conf->chunk_time_given)
		pord->chunk_tv.tv_usec = pord->conf->chunk_time_arg;
	else
		pord->chunk_tv.tv_usec = rn->chunk_time;
	pord->chunk_ts = rtp_session_time_to_ts(pord->session, tv2ms(&pord->chunk_tv));
	PARA_INFO_LOG("chunk time: %lums\n", tv2ms(&pord->chunk_tv));
	PARA_NOTICE_LOG("receiving from %s:%d\n", pord->conf->host_arg, pord->conf->port_arg);
	rtp_session_set_local_addr(pord->session, pord->conf->host_arg, pord->conf->port_arg);
	rtp_session_set_payload_type(pord->session, PAYLOAD_AUDIO_CONTINUOUS);
	return 1;
}

int ortp_recv_init(struct receiver *r)
{
	r->shutdown = ortp_shutdown;
	r->open = ortp_recv_open;
	r->close = ortp_recv_close;
	r->pre_select = ortp_recv_pre_select;
	r->post_select = ortp_recv_post_select;
	r->print_help = ortp_recv_cmdline_parser_print_help;
	r->parse_config = ortp_recv_parse_config;

	ortp_init();
	ortp_set_debug_file("oRTP", NULL);
	return 1;
}
